#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.core.solvers import get_solver
from idaes_flowsheet_processor.api import FlowsheetInterface
from watertap.flowsheets.lsrro.lsrro import (
    build,
    set_operating_conditions,
    initialize,
    optimize_set_up,
    solve,
    ACase,
    BCase,
    ABTradeoff,
)
from pyomo.environ import units as pyunits


def export_to_ui():
    return FlowsheetInterface(
        name="LSRRO",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
        build_options={
            "NumberOfStages": {
                "name": "NumberOfStages",
                "display_name": "Number of stages",
                "values_allowed": "int",
                "value": 3,  # default value
                "max_val": 8,  # optional
                "min_val": 1,  # optional
            },
            "FeedFlowRate": {
                "name": "FeedFlowRate",
                "display_name": "Feed Flow Rate (m3/s)",
                "values_allowed": "float",
                "value": 1e-3,  # default value
                "max_val": 1,  # optional
                "min_val": 1e-5,  # optional
            },
            "FeedNaClConcentration": {
                "name": "FeedNaClConcentration",
                "display_name": "Feed NaCl Concentration (kg/m3)",
                "values_allowed": "float",
                "value": 70,  # default value
                "max_val": 100,  # optional
                "min_val": 1,  # optional
            },
            "ROFiniteElements": {
                "name": "ROFiniteElements",
                "display_name": "RO Finite Elements",
                "values_allowed": "float",
                "value": 10,  # default value
                "max_val": 20,  # optional
                "min_val": 1,  # optional
            },
            "NaClSolubilityLimit": {
                "name": "NaClSolubilityLimit",
                "display_name": "NaCl Solubility Limit",
                "values_allowed": ["False", "True"],
                "value": "True",  # default value
            },
            "ConcentrationPolarization": {
                "name": "ConcentrationPolarization",
                "display_name": "Calculate Concentration Polzarization",
                "values_allowed": ["False", "True"],
                "value": "True",  # default value
            },
            "ROPressureDrop": {
                "name": "ROPressureDrop",
                "display_name": "Calculate Pressure Drop",
                "values_allowed": ["False", "True"],
                "value": "True",  # default value
            },
            "ACase": {
                "name": "ACase",
                "display_name": "Water Permeability",
                "values_allowed": [
                    x.name for x in [ACase.optimize, ACase.fixed, ACase.single_optimum]
                ],
                "value": ACase.fixed.name,  # default value
            },
            "BCase": {
                "name": "BCase",
                "display_name": "Salt Permeability",
                "values_allowed": [
                    x.name for x in [BCase.optimize, BCase.single_optimum]
                ],
                "value": BCase.optimize.name,  # default value
            },
            "ABTradeoff": {
                "name": "ABTradeoff",
                "display_name": "Water & Salt Permeability Equality Constraints",
                "values_allowed": [
                    x.name
                    for x in [
                        ABTradeoff.inequality_constraint,
                        ABTradeoff.equality_constraint,
                        ABTradeoff.none,
                    ]
                ],
                "value": ABTradeoff.none.name,  # default value
            },
            "BMax": {
                "name": "BMax",
                "display_name": "Maximum NaCl Permeability",
                "values_allowed": "float",
                "value": 3.5e-6,  # default value
                "max_val": 1e-5,  # optional
                "min_val": 1e-7,  # optional
            },
        },
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    exports.add(
        obj=fs.feed.properties[0].flow_vol_phase["Liq"],
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hour,
        display_units="m3/hr",
        rounding=2,
        description="Inlet volumetric flow rate",
        is_input=False,
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.properties[0].conc_mass_phase_comp["Liq", "NaCl"],
        name="NaCl concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=2,
        description="Inlet salinity",
        is_input=False,
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, OARO
    for idx, pump in fs.PrimaryPumps.items():
        exports.add(
            obj=pump.control_volume.properties_out[0].pressure,
            name=f"Pump {idx} outlet pressure",
            ui_units=pyunits.bar,
            display_units="bar",
            rounding=1,
            description=f"Outlet pressure of pump {idx}",
            is_input=True,
            input_category="Primary Pumps",
            is_output=False,
        )
        exports.add(
            obj=pump.efficiency_pump[0],
            name=f"Pump {idx} efficiency",
            ui_units=pyunits.dimensionless,
            display_units="fraction",
            rounding=2,
            description=f"Efficiency of pump {idx}",
            is_input=True,
            input_category="Primary Pumps",
            is_output=False,
        )
    for idx, pump in fs.BoosterPumps.items():
        exports.add(
            obj=pump.efficiency_pump[0],
            name=f"Pump {idx} efficiency",
            ui_units=pyunits.dimensionless,
            display_units="fraction",
            rounding=2,
            description=f"Efficiency of pump {idx}",
            is_input=True,
            input_category="Booster Pumps",
            is_output=False,
        )
    for idx, stage in fs.ROUnits.items():
        name = f"Stage {idx}"
        exports.add(
            obj=stage.A_comp[0, "H2O"],
            name=f"Stage {idx} water permeability coefficient",
            ui_units=pyunits.L / pyunits.hr / pyunits.m**2 / pyunits.bar,
            display_units="LMH/bar",
            rounding=2,
            description=f"Membrane water permeability",
            is_input=True,
            input_category=f"{name}",
            is_output=False,
        )
        exports.add(
            obj=stage.B_comp[0, "NaCl"],
            name=f"Stage {idx} salt permeability coefficient",
            ui_units=pyunits.L / pyunits.hr / pyunits.m**2,
            display_units="LMH",
            rounding=2,
            description=f"Membrane salt permeability",
            is_input=True,
            input_category=f"{name}",
            is_output=False,
        )
        exports.add(
            obj=stage.area,
            name=f"Stage {idx} membrane area",
            ui_units=pyunits.m**2,
            display_units="m2",
            rounding=2,
            description=f"Membrane area",
            is_input=True,
            input_category=f"{name}",
            is_output=True,
            output_category="Membrane area",
        )
        exports.add(
            obj=stage.width,
            name=f"Stage {idx} width",
            ui_units=pyunits.m,
            display_units="m",
            rounding=2,
            description=f"Membrane width",
            is_input=True,
            input_category=f"{name}",
            is_output=False,
        )
        exports.add(
            obj=stage.mixed_permeate[0].pressure,
            name=f"Stage {idx} permeate pressure",
            ui_units=pyunits.bar,
            display_units="bar",
            rounding=1,
            description=f"Membrane permeate pressure",
            is_input=True,
            input_category=f"{name}",
            is_output=False,
        )
        if (
            build_options["ROPressureDrop"]
            or build_options["ConcentrationPolarization"]
        ):
            exports.add(
                obj=stage.feed_side.channel_height,
                name=f"Stage {idx} channel height",
                ui_units=pyunits.mm,
                display_units="mm",
                rounding=1,
                description=f"Channel height in membrane stage",
                is_input=True,
                input_category=f"{name}",
                is_output=False,
            )
            exports.add(
                obj=stage.feed_side.spacer_porosity,
                name=f"Stage {idx} space porosity",
                ui_units=pyunits.dimensionless,
                display_units="fraction",
                rounding=2,
                description=f"Spacer porosity in membrane stage",
                is_input=True,
                input_category=f"{name}",
                is_output=False,
            )
    for idx, erd in fs.EnergyRecoveryDevices.items():
        exports.add(
            obj=erd.efficiency_pump[0],
            name=f"ERD {idx} efficiency",
            ui_units=pyunits.dimensionless,
            display_units="fraction",
            rounding=2,
            description=f"Energy recovery device {idx} efficiency",
            is_input=True,
            input_category="Energy Recovery Device",
            is_output=False,
        )

    # System costing
    exports.add(
        obj=fs.costing.utilization_factor,
        name="Utilization factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Utilization factor - [annual use hours/total hours in year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.TIC,
        name="Total Installed Cost",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=1,
        description="Total Installed Cost (TIC)",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.maintenance_labor_chemical_factor,
        name="Maintenance-labor-chemical factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Maintenance-labor-chemical factor [fraction of investment cost/year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.capital_recovery_factor,
        name="Capital annualization factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Capital annualization factor [fraction of investment cost/year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kWh",
        rounding=3,
        description="Electricity cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.reverse_osmosis.factor_membrane_replacement,
        name="Membrane replacement factor",
        ui_units=pyunits.yr**-1,
        display_units="1/yr",
        rounding=3,
        description="Membrane replacement factor [fraction of membrane replaced/year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.reverse_osmosis.membrane_cost,
        name="Membrane cost",
        ui_units=pyunits.USD_2018 / (pyunits.m**2),
        display_units="$/m2",
        rounding=3,
        description="Membrane cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.reverse_osmosis.high_pressure_membrane_cost,
        name="High pressure membrane cost",
        ui_units=pyunits.USD_2018 / (pyunits.m**2),
        display_units="$/m2",
        rounding=3,
        description="High pressure membrane cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.high_pressure_pump.cost,
        name="High pressure pump cost",
        ui_units=pyunits.USD_2018 / pyunits.watt,
        display_units="$/W",
        rounding=3,
        description="High pressure pump cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.energy_recovery_device.pressure_exchanger_cost,
        name="ERD pressure exchanger cost",
        ui_units=pyunits.USD_2018 / (pyunits.meter**3 / pyunits.hours),
        display_units="$/m3/hr",
        rounding=3,
        description="Pressure exchanger cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )

    # System constraints
    exports.add(
        obj=fs.water_recovery,
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Water recovery",
        is_input=True,
        input_category="System constraints",
        is_output=True,
        output_category="System constraints",
    )
    exports.add(
        obj=fs.ro_max_pressure,
        name="Maximum allowable RO pressure",
        ui_units=pyunits.bar,
        display_units="bar",
        rounding=2,
        description="Maximum allowable pressure for RO",
        is_input=True,
        input_category="System constraints",
        is_output=True,
        output_category="System constraints",
    )
    exports.add(
        obj=fs.lsrro_max_pressure,
        name="Maximum allowable LSRRO pressure",
        ui_units=pyunits.bar,
        display_units="bar",
        rounding=2,
        description="Maximum allowable pressure for LSRRO",
        is_input=True,
        input_category="System constraints",
        is_output=True,
        output_category="System constraints",
    )

    # --- Output data ---
    # Product
    exports.add(
        obj=fs.product.properties[0].flow_vol,
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet product water volumetric flow rate",
        is_input=False,
        is_output=True,
        output_category="Product",
    )
    exports.add(
        obj=fs.product.properties[0].conc_mass_phase_comp["Liq", "NaCl"],
        name="NaCl concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=3,
        description="Outlet product water NaCl concentration",
        is_input=False,
        is_output=True,
        output_category="Product",
    )

    # Disposal
    exports.add(
        obj=fs.disposal.properties[0].flow_vol,
        name="Volumetric flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet product water volumetric flow rate",
        is_input=False,
        is_output=True,
        output_category="Disposal",
    )
    exports.add(
        obj=fs.disposal.properties[0].conc_mass_phase_comp["Liq", "NaCl"],
        name="NaCl concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=3,
        description="Outlet product water NaCl concentration",
        is_input=False,
        is_output=True,
        output_category="Disposal",
    )
    # Primary Pumps
    exports.add(
        obj=fs.total_pump_work,
        name="Pump work",
        ui_units=pyunits.kW,
        display_units="kW",
        rounding=3,
        description="Total work by primary pumps",
        is_input=False,
        is_output=True,
        output_category="Primary Pumps",
    )
    exports.add(
        obj=fs.recovered_pump_work,
        name="Pump work recovered",
        ui_units=pyunits.kW,
        display_units="kW",
        rounding=3,
        description="Total work recovered by primary pumps",
        is_input=False,
        is_output=True,
        output_category="Primary Pumps",
    )
    exports.add(
        obj=fs.net_pump_work,
        name="Net pump work",
        ui_units=pyunits.kW,
        display_units="kW",
        rounding=3,
        description="Net pump work by primary pumps",
        is_input=False,
        is_output=True,
        output_category="Primary Pumps",
    )
    energy_recovery = -fs.recovered_pump_work / fs.total_pump_work * 100
    exports.add(
        obj=energy_recovery,
        name="Energy recovery",
        ui_units=pyunits.dimensionless,
        display_units="%",
        rounding=3,
        description="Energy recovery by primary pumps",
        is_input=False,
        is_output=True,
        output_category="Primary Pumps",
    )

    # System metrics
    exports.add(
        obj=fs.NumberOfStages,
        name="Number of stages",
        ui_units=pyunits.dimensionless,
        display_units=" ",
        rounding=0,
        description="Number of stages",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.water_recovery * 100,
        name="Water volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="%",
        rounding=2,
        description="System water volumetric recovery",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.mass_water_recovery * 100,
        name="Mass water recovery rate",
        ui_units=pyunits.dimensionless,
        display_units="%",
        rounding=2,
        description="System water mass recovery",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.system_salt_rejection * 100,
        name="Salt rejection",
        ui_units=pyunits.dimensionless,
        display_units="%",
        rounding=2,
        description="System salt rejection",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.costing.specific_energy_consumption,
        name="Specific energy consumption",
        ui_units=pyunits.kWh / pyunits.m**3,
        display_units="kWh/m3 of product water",
        rounding=3,
        description="Specific energy consumption (SEC)",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    electricity_cost = (
        fs.costing.aggregate_flow_costs["electricity"]
        * fs.costing.utilization_factor
        / fs.costing.annual_water_production
    )
    exports.add(
        obj=electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Electricity cost",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    total_area = sum(fs.ROUnits[i].area for i in fs.NonFinalStages)
    exports.add(
        obj=total_area,
        name="Total membrane area",
        ui_units=pyunits.m**2,
        display_units="m^2",
        rounding=2,
        description="Total membrane area",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.annual_feed,
        name="Annual feed flow",
        ui_units=pyunits.m**3 / pyunits.yr,
        display_units="m3/yr",
        rounding=3,
        description="Annual feed water flow rate",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )
    exports.add(
        obj=fs.costing.annual_water_production,
        name="Annual water production",
        ui_units=pyunits.m**3 / pyunits.yr,
        display_units="m3/yr",
        rounding=3,
        description="Annual feed water production",
        is_input=False,
        is_output=True,
        output_category="System metrics",
    )

    exports.add(
        obj=fs.costing.LCOW,
        name="Levelized cost of water",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Levelized cost of water (LCOW)",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    total_capex = (
        fs.costing.total_capital_cost
        * fs.costing.capital_recovery_factor
        / fs.costing.annual_water_production
    )
    exports.add(
        obj=total_capex,
        name="Total CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Total capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.primary_pump_capex_lcow,
        name="Primary pump CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Primary pump capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.booster_pump_capex_lcow,
        name="Booster pump CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Booster pump capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.erd_capex_lcow,
        name="Energy recovery device CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="ERD capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.membrane_capex_lcow,
        name="Membrane CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Membrane capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.indirect_capex_lcow,
        name="Indirect CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Indirect capital expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.electricity_lcow,
        name="Electricity LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Electricity levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.membrane_replacement_lcow,
        name="Membrane replacement LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Membrane replacement levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.chemical_labor_maintenance_lcow,
        name="Chemical-labor-maintenance CAPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="CLM levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    exports.add(
        obj=fs.costing.pumping_energy_aggregate_lcow,
        name="Pumping energy aggregate LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="System pumping energy levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )
    total_opex = fs.costing.total_operating_cost / fs.costing.annual_water_production
    exports.add(
        obj=total_opex,
        name="Total OPEX LCOW",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product water",
        rounding=3,
        description="Total operating expenses levelized cost of water",
        is_input=False,
        is_output=True,
        output_category="LCOW breakdown",
    )

    # Stage metrics
    for idx, pump in fs.PrimaryPumps.items():
        exports.add(
            obj=pyunits.convert(
                pump.control_volume.properties_out[0].pressure, to_units=pyunits.bar
            ),
            name=f"Feed pressure - stage {idx}",
            ui_units=pyunits.bar,
            display_units="bar",
            rounding=1,
            description=f"Feed pressure of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="Primary pump feed pressure",
        )

    for idx, stage in fs.ROUnits.items():
        exports.add(
            obj=stage.rejection_phase_comp[0, "Liq", "NaCl"] * 100,
            name=f"Observed NaCl rejection - stage {idx}",
            ui_units=pyunits.dimensionless,
            display_units="%",
            rounding=2,
            description=f"Observed salt rejection of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO observed rejection",
        )

    for idx, stage in fs.ROUnits.items():
        A_comp = pyunits.convert(
            stage.A_comp[0, "H2O"],
            to_units=pyunits.L / pyunits.hr / pyunits.m**2 / pyunits.bar,
        )
        exports.add(
            obj=A_comp,
            name=f"Water permeability coefficient - stage {idx}",
            ui_units=pyunits.L / pyunits.hr / pyunits.m**2 / pyunits.bar,
            display_units="LMH/bar",
            rounding=2,
            description=f"A value of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO water permeability coefficient (A)",
        )

    for idx, stage in fs.ROUnits.items():
        B_comp = pyunits.convert(
            stage.B_comp[0, "NaCl"], to_units=pyunits.L / pyunits.hr / pyunits.m**2
        )
        exports.add(
            obj=B_comp,
            name=f"Salt permeability coefficient - stage {idx}",
            ui_units=pyunits.L / pyunits.hr / pyunits.m**2,
            display_units="LMH",
            rounding=2,
            description=f"B value of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO salt permeability coefficient (A)",
        )

    for idx, stage in fs.ROUnits.items():
        water_flux = pyunits.convert(
            stage.flux_mass_phase_comp_avg[0, "Liq", "H2O"],
            to_units=pyunits.kg / pyunits.hr / pyunits.m**2,
        )
        exports.add(
            obj=water_flux,
            name=f"Average water flux - stage {idx}",
            ui_units=pyunits.kg / pyunits.hr / pyunits.m**2,
            display_units="kg/hr-m2",
            rounding=2,
            description=f"Average water flux of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO water flux",
        )

    for idx, stage in fs.ROUnits.items():
        salt_flux = pyunits.convert(
            stage.flux_mass_phase_comp_avg[0, "Liq", "NaCl"],
            to_units=pyunits.g / pyunits.m**2 / pyunits.h,
        )
        exports.add(
            obj=salt_flux,
            name=f"Average NaCl flux - stage {idx}",
            ui_units=pyunits.g / pyunits.m**2 / pyunits.h,
            display_units="gMH",
            rounding=1,
            description=f"Average salt flux of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO NaCl flux",
        )

    for idx, stage in fs.ROUnits.items():
        exports.add(
            obj=stage.feed_side.N_Re[0, 0],
            name=f"Inlet Reynolds number - stage {idx}",
            ui_units=pyunits.dimensionless,
            display_units=" ",
            rounding=2,
            description=f"Reynolds number of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO Reynolds number",
        )

    for idx, stage in fs.ROUnits.items():
        exports.add(
            obj=stage.feed_side.velocity[0, 0],
            name=f"Inlet crossflow velocity - stage {idx}",
            ui_units=pyunits.m / pyunits.s,
            display_units="m/s",
            rounding=2,
            description=f"Inlet velocity of stage {idx}",
            is_input=False,
            is_output=True,
            output_category="RO Crossflow velocity",
        )


def build_flowsheet(build_options=None, **kwargs):
    if build_options is not None:
        # get solver
        solver = get_solver()

        # build, set, and initialize
        m = build(
            number_of_stages=build_options["NumberOfStages"].value,
            has_NaCl_solubility_limit=build_options["NaClSolubilityLimit"].value,
            has_calculated_concentration_polarization=build_options[
                "ConcentrationPolarization"
            ].value,
            has_calculated_ro_pressure_drop=build_options["ROPressureDrop"].value,
            number_of_RO_finite_elements=build_options["ROFiniteElements"].value,
            B_max=build_options["BMax"].value,
        )
        set_operating_conditions(
            m,
            Cin=build_options["FeedNaClConcentration"].value,
            Qin=build_options["FeedFlowRate"].value,
        )

        initialize(m)

        solve(m, solver=solver)

        # Configuration options taken from user settings and run_lsrro_case
        if build_options["ACase"].value == ACase.fixed:
            optimize_set_up(
                m,
                set_default_bounds_on_module_dimensions=True,
                Cbrine=None,
                A_case=build_options["ACase"].value,
                B_case=build_options["BCase"].value,
                AB_tradeoff=build_options["ABTradeoff"].value,
                A_value=4.2e-12,
                permeate_quality_limit=500e-6,
                AB_gamma_factor=1,
                B_max=build_options["BMax"].value,
            )
        else:
            optimize_set_up(
                m,
                set_default_bounds_on_module_dimensions=True,
                Cbrine=None,
                A_case=build_options["ACase"].value,
                B_case=build_options["BCase"].value,
                AB_tradeoff=build_options["ABTradeoff"].value,
                permeate_quality_limit=500e-6,
                AB_gamma_factor=1,
                B_max=build_options["BMax"].value,
            )

        solve(m, raise_on_failure=False, tee=False, solver=solver)

    else:
        # get solver
        solver = get_solver()

        # build, set, and initialize
        m = build(
            number_of_stages=3,
            has_NaCl_solubility_limit=True,
            has_calculated_concentration_polarization=True,
            has_calculated_ro_pressure_drop=True,
            number_of_RO_finite_elements=10,
            B_max=None,
        )
        set_operating_conditions(m, Cin=None, Qin=None)

        initialize(m)
        solve(m, solver=solver)

        # Configuration options taken from def optimize_set_up
        optimize_set_up(
            m,
            set_default_bounds_on_module_dimensions=True,
            water_recovery=None,
            Cbrine=None,
            A_case=ACase.fixed,
            A_value=4.2e-12,
            B_case=BCase.optimize,
            AB_tradeoff=ABTradeoff.none,
            permeate_quality_limit=None,
            AB_gamma_factor=None,
            B_max=None,
        )

        # display
        solve(m, solver=solver)

    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from watertap.unit_models.generic_desalter import (
    GenericDesalter,
)
from pyomo.environ import (
    value,
)
from watertap.flowsheets.generic_desalination_train.costing import (
    desalter_costing,
)
import logging

_logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "desalter %(asctime)s %(levelname)s: %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)


__author__ = "Alexander V. Dudchenko"


def build(
    m,
    block,
    base_cost=1,
    recovery_cost=0,
    tracked_solids=None,
    min_recovery=0,
    max_recovery=99,
    default_recovery=80,
):
    block.desalter = GenericDesalter(
        property_package=m.fs.properties, tracked_solids_list=tracked_solids
    )

    block.desalter.water_recovery.setub(max_recovery)
    block.desalter.water_recovery.setlb(min_recovery)

    block.desalter.water_recovery.fix(default_recovery)
    desalter_costing.cost_desalter(
        m.fs.costing,
        block.desalter,
        base_cost,
        recovery_cost,
        opt_name=block.process_name,
    )

    block.desalter.brine_unit.properties_out[0].flow_vol_phase[...]


def initialize(m, blk, solver):
    blk.desalter.initialize(optarg=solver.options)


def unfix_opt_vars(m, blk):
    blk.desalter.water_recovery.unfix()


def display(m, blk):
    _logger.info(
        f"Feed flow {value(blk.desalter.brine_unit.properties_in[0].flow_mass_phase_comp['Liq','H2O'])}"
    )
    _logger.info(
        f"Brine flow {value(blk.desalter.brine_unit.properties_out[0].flow_mass_phase_comp['Liq','H2O'])}"
    )
    _logger.info(
        f"Product flow {value(blk.desalter.product_properties[0].flow_mass_phase_comp['Liq','H2O'])}"
    )
    _logger.info(
        f"Brine water content (%) {value(blk.desalter.brine_water_mass_percent)}"
    )
    _logger.info(f"Recovery (%) {value(blk.desalter.water_recovery)}")
    _logger.info(f"Annual cost ($) {value(blk.desalter.annual_cost)}")



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    Objective,
    Block,
)
from pyomo.network import Port

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import Feed, Product, Separator
from watertap.unit_models.electrodialysis_1D import (
    ElectricalOperationMode,
    Electrodialysis1D,
)

from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock

import watertap.flowsheets.electrodialysis.electrodialysis_1stack as edfs

__author__ = "Xiangyu Bi"


class TestElectrodialysisVoltageConst:
    @pytest.fixture(scope="class")
    def electrodialysis_1D1stack(self):
        m = edfs.build()
        return m

    @pytest.mark.unit
    def test_build_model(self, electrodialysis_1D1stack):
        m = electrodialysis_1D1stack

        # Test basic build
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, MCASParameterBlock)
        assert isinstance(m.fs.costing, Block)
        assert isinstance(m.fs.feed, Feed)
        assert isinstance(m.fs.separator, Separator)
        assert isinstance(m.fs.EDstack, Electrodialysis1D)
        assert isinstance(m.fs.product, Product)
        assert isinstance(m.fs.disposal, Product)

        # Test port
        assert isinstance(m.fs.feed.outlet, Port)
        assert isinstance(m.fs.separator.inlet, Port)
        assert isinstance(m.fs.separator.inlet_diluate, Port)
        assert isinstance(m.fs.separator.inlet_concentrate, Port)
        assert isinstance(m.fs.EDstack.inlet_diluate, Port)
        assert isinstance(m.fs.EDstack.inlet_concentrate, Port)
        assert isinstance(m.fs.product.inlet, Port)
        assert isinstance(m.fs.disposal.inlet, Port)

        # Test consting
        assert isinstance(m.fs.EDstack.costing, Block)
        assert isinstance(m.fs.EDstack.costing.capital_cost, Var)
        assert isinstance(m.fs.EDstack.costing.fixed_operating_cost, Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(m.fs.costing, var_str)
            assert isinstance(var, Var)

        # Test arcs
        arc_dict = {
            m.fs.s01: (m.fs.feed.outlet, m.fs.separator.inlet),
            m.fs.s02: (m.fs.separator.inlet_diluate, m.fs.EDstack.inlet_diluate),
            m.fs.s03: (
                m.fs.separator.inlet_concentrate,
                m.fs.EDstack.inlet_concentrate,
            ),
            m.fs.s04: (m.fs.EDstack.outlet_diluate, m.fs.product.inlet),
            m.fs.s05: (m.fs.EDstack.outlet_concentrate, m.fs.disposal.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

        # Test the primary EDstack properties
        # test configrations
        assert len(m.fs.EDstack.config) == 21
        assert not m.fs.EDstack.config.dynamic
        assert not m.fs.EDstack.config.has_holdup
        assert (
            m.fs.EDstack.config.operation_mode
            == ElectricalOperationMode.Constant_Voltage
        )
        assert (
            m.fs.EDstack.config.material_balance_type == MaterialBalanceType.useDefault
        )
        assert (
            m.fs.EDstack.config.momentum_balance_type
            == MomentumBalanceType.pressureTotal
        )
        assert m.fs.EDstack.config.property_package is m.fs.properties
        assert "H2O" in m.fs.properties.component_list

    @pytest.mark.component
    def test_specific_operating_conditions(self, electrodialysis_1D1stack):
        m = electrodialysis_1D1stack
        edfs.set_operating_conditions(m)
        assert degrees_of_freedom(m) == 0
        edfs.initialize_system(m)
        edfs.solve(m)
        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            8.7e-5, abs=1e-6
        )
        assert value(m.fs.product.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            4.2e-5, abs=1e-6
        )
        assert value(
            m.fs.disposal.properties[0].flow_vol_phase["Liq"]
        ) == pytest.approx(4.5e-5, abs=1e-6)
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(9.895, rel=1e-3)
        assert value(
            sum(
                m.fs.product.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(4.479, rel=1e-3)
        assert value(
            sum(
                m.fs.disposal.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(14.942, rel=1e-3)

        assert value(m.fs.EDstack.recovery_mass_H2O[0]) == pytest.approx(
            0.485, rel=1e-3
        )
        assert value(
            m.fs.EDstack.cell_width
            * m.fs.EDstack.cell_length
            * m.fs.EDstack.cell_pair_num
        ) == pytest.approx(7.900, rel=1e-3)
        assert value(m.fs.EDstack.voltage_applied[0]) == 5
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            0.1973, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(1.1142, rel=1e-3)

    @pytest.mark.component
    def test_optimization(self, electrodialysis_1D1stack):
        m = electrodialysis_1D1stack
        edfs.initialize_system(m)
        edfs.optimize_system(m)
        isinstance(m.fs.objective, Objective)
        assert m.fs.objective.expr == m.fs.costing.LCOW

        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            8.7e-5, abs=1e-6
        )
        assert value(m.fs.product.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            4.2e-5, abs=1e-6
        )
        assert value(
            m.fs.disposal.properties[0].flow_vol_phase["Liq"]
        ) == pytest.approx(4.5e-5, abs=1e-6)
        assert value(m.fs.product_salinity) == pytest.approx(0.9996, rel=1e-3)
        assert value(m.fs.disposal_salinity) == pytest.approx(18.1124, rel=1e-3)

        assert value(m.fs.EDstack.recovery_mass_H2O[0]) == pytest.approx(
            0.48455, rel=1e-3
        )
        assert value(m.fs.mem_area) == pytest.approx(1.1060, rel=1e-3)
        assert value(m.fs.EDstack.voltage_applied[0]) == pytest.approx(
            7.03676, rel=1e-3
        )
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            2.2922, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.42547, rel=1e-3)

    @pytest.mark.unit
    def test_main_fun(self, electrodialysis_1D1stack):
        edfs.main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from .complete_condenser import Condenser
from .compressor import Compressor
from .evaporator import Evaporator



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    Constraint,
    Objective,
    TransformationFactory,
    NonNegativeReals,
    units as pyunits,
)
from idaes.core import (
    FlowsheetBlock,
)

import idaes.core.util.scaling as iscale
from idaes.core.util.model_statistics import degrees_of_freedom


from watertap.core.solvers import get_solver

from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    ActivityCoefficientModel,
    DensityCalculation,
    MaterialFlowBasis,
)

from idaes.core.util.initialization import propagate_state

from idaes.models.unit_models import (
    Product,
    Feed,
)

import watertap.flowsheets.generic_desalination_train.utils.scale_utils as scaleTools
from pyomo.util.check_units import assert_units_consistent
from watertap.flowsheets.generic_desalination_train.costing import (
    generic_costing,
    stream_costing,
)
from pyomo.network import Arc
from watertap.flowsheets.generic_desalination_train.unit_operations import (
    desalter,
    mixer,
    separator,
)

import logging

_logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "gen_train %(asctime)s %(levelname)s: %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)

__author__ = "Alexander V. Dudchenko"


def main():
    m = build()
    initialize(m)
    m.fs.Pretreatment.separator.component_removal_percent["X"].fix(50)
    m.fs.Pretreatment.separator.separation_cost["X"].fix(0.5)
    m.fs.Valorizer.separator.product_value["X"].fix(1)
    m.fs.Valorizer.separator.component_removal_percent["X"].fix(50)

    m.fs.Desal_1.desalter.water_recovery.fix(80)
    m.fs.Desal_2.desalter.water_recovery.fix(50)
    m.fs.Desal_2.desalter.recovery_cost.fix(0.01)
    m.fs.Desal_2.desalter.recovery_cost_offset.fix(35)
    m.fs.Desal_3.desalter.water_recovery.unfix()
    m.fs.Desal_3.desalter.brine_water_mass_percent.fix(80)
    result = solve(m)
    return result


def build(
    train_type="Pretreatment>Desal1>Desal2>Crystalizer>Valorizer",
    water_source="generic",
):
    train_orders = {
        "Pretreatment>Desal1>Desal2>Crystalizer>Valorizer": {
            0: {
                "process_type": "separator",
                "process_name": "Pretreatment",
                "default_kwargs": {
                    "base_cost": 0.1,
                    "additive_dose": 0,
                    "additive_cost": 0,
                },
            },
            1: {
                "process_type": "desalter",
                "process_name": "Desal_1",
                "default_kwargs": {
                    "base_cost": 0.3,
                    "recovery_cost": 0,
                    "min_recovery": 0,
                    "max_recovery": 85,
                    "default_recovery": 85,
                },
            },
            2: {
                "process_type": "desalter",
                "process_name": "Desal_2",
                "default_kwargs": {
                    "base_cost": 0.5,
                    "recovery_cost": 0,
                    "min_recovery": 0,
                    "max_recovery": 50,
                    "default_recovery": 50,
                },
            },
            3: {
                "process_type": "desalter",
                "process_name": "Desal_3",
                "default_kwargs": {
                    "base_cost": 10,
                    "recovery_cost": 0.0,
                    "min_recovery": 0,
                    "max_recovery": 99.9,
                    "default_recovery": 90,
                },
            },
            4: {
                "process_type": "valorizer",
                "process_name": "Valorizer",
                "default_kwargs": {
                    "base_cost": 0,
                    "additive_dose": 0.0,
                    "additive_cost": 0.0,
                },
            },
        },
    }

    train_order = train_orders[train_type]
    """ set up mcas as our default prop package for tracking ions etc."""

    mcas_props = {
        "activity_coefficient_model": ActivityCoefficientModel.ideal,
        "density_calculation": DensityCalculation.constant,
        "solute_list": ["TDS", "X"],
        "mw_data": {"H2O": 0.01801528, "TDS": 1, "X": 1},
        "charge": {"TDS": 0.0, "X": 0.0},
        "material_flow_basis": MaterialFlowBasis.mass,
    }

    # defines solids used for precipitator tracking

    tracked_solids = ["TDS"]
    """ build all the processes and connection"""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(**mcas_props)
    m.fs.costing = generic_costing.add_generic_costing_block()

    m.working_name = water_source + "_" + train_type.replace(">", "_")

    m.fs.feed = Feed(property_package=m.fs.properties)
    """  make sure these are constructed."""
    m.fs.feed.properties[0].conc_mass_phase_comp[...]
    m.fs.feed.properties[0].flow_mass_phase_comp[...]
    m.fs.feed.properties[0].flow_vol_phase[...]

    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    stream_costing.cost_stream(
        m.fs.costing, m.fs.product, 0, opt_name="Product distribution"
    )
    stream_costing.cost_stream(
        m.fs.costing, m.fs.disposal, 1, opt_name="Waste disposal"
    )
    stream_costing.cost_stream(m.fs.costing, m.fs.feed, 0, opt_name="Feed sourcing")

    add_flowsheet_level_constraints(m, m.fs)

    """ build main processes and arcs"""

    m.fs.process_order = []
    m.arc_order = []
    m.end_point_order = []
    for item_number in range(len(train_order.keys())):
        process_name = train_order[item_number]["process_type"]
        if train_order[item_number].get("default_kwargs") is None:
            train_order[item_number]["default_kwargs"] = {}
        else:
            train_order[item_number]["default_kwargs"].update(
                {"tracked_solids": tracked_solids}
            )
        build_selected_processes(m, **train_order[item_number])
        if item_number == 0:
            arc_name = "Feed_to_{}".format(process_name)
            m.fs.add_component(
                arc_name,
                Arc(
                    source=m.fs.feed.outlet,
                    destination=m.fs.process_order[-1]["feed"],
                ),
            )
            m.arc_order.append(m.fs.find_component(arc_name))
        else:
            prior_process_name = m.fs.process_order[-2]["process_name"]
            arc_name = "{}_to_{}".format(prior_process_name, process_name)
            m.fs.add_component(
                arc_name,
                Arc(
                    source=m.fs.process_order[-2]["effluent"],
                    destination=m.fs.process_order[-1]["feed"],
                ),
            )
            m.arc_order.append(m.fs.find_component(arc_name))

    product_streams = []
    disposal_streams = []
    for i, process_dict in enumerate(m.fs.process_order):
        if process_dict.get("product") is not None:
            product_streams.append(process_dict)
        if process_dict.get("disposal") is not None or i == len(m.fs.process_order) - 1:
            if (
                process_dict.get("disposal") is None
                and process_dict.get("effluent") is not None
            ):
                process_dict["disposal"] = process_dict["effluent"]
            if process_dict.get("disposal") is not None:
                disposal_streams.append(process_dict)
    streams = {"product": product_streams, "disposal": disposal_streams}
    for end_point, streams in streams.items():
        if len(streams) > 1:
            mixer.build_mixer(
                m,
                "{}_mixer".format(end_point),
                [process_dict["process_name"] for process_dict in streams],
            )
            mixer_unit = m.fs.find_component("{}_mixer".format(end_point))
            for stream in streams:
                arc_name = "{}_to_{}_mixer".format(stream["process_name"], end_point)
                m.fs.add_component(
                    arc_name,
                    Arc(
                        source=stream[end_point],
                        destination=mixer_unit.find_component(stream["process_name"]),
                    ),
                )
                m.end_point_order.append(m.fs.find_component(arc_name))
            arc_name = "{}_mixer_to_{}".format(end_point, end_point)
            m.fs.add_component(
                arc_name,
                Arc(
                    source=mixer_unit.outlet,
                    destination=m.fs.find_component(end_point).inlet,
                ),
            )
            m.end_point_order.append(m.fs.find_component(arc_name))
        else:
            arc_name = "{}_to_{}".format(streams[0]["process_name"], end_point)
            m.fs.add_component(
                arc_name,
                Arc(
                    source=streams[0][end_point],
                    destination=m.fs.find_component(end_point).inlet,
                ),
            )
            m.end_point_order.append(m.fs.find_component(arc_name))

    generic_costing.cost_process(
        m.fs.costing, m.fs.product.properties[0].flow_vol_phase["Liq"]
    )

    TransformationFactory("network.expand_arcs").apply_to(m)

    """set in initial operating conditions"""
    set_feed(m)

    """ set scaling for feed_composition"""
    scaleTools.set_default_scaling(
        m.fs.feed.properties[0],
        m.fs.properties,
        "flow_mol_phase_comp",
        factor=1.0,
    )
    scaleTools.set_default_scaling(
        m.fs.feed.properties[0],
        m.fs.properties,
        "flow_mass_phase_comp",
        factor=1.0,
    )
    m.fs.cost_objective = Objective(expr=m.fs.costing.LCOW)
    assert_units_consistent(m)
    _logger.info(f"DOFS:{degrees_of_freedom(m)}")
    assert degrees_of_freedom(m) == 0
    return m


def set_feed(m):
    m.fs.feed.properties[0].temperature.fix(298.15)
    m.fs.feed.properties[0].pressure.fix(101325)
    m.fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix(1)
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "TDS"].fix(3.5)  # kg/m3 or g/L
    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", "X"].fix(0.3)  # kg/m3 or g/L
    update_feed(m.fs)


def add_flowsheet_level_constraints(m, blk):
    blk.water_recovery = Var(
        initialize=50,
        bounds=(0, 99.5),
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        doc="System Water Recovery",
    )
    blk.water_recovery.setub(99.5)
    blk.water_recovery.setlb(0)
    blk.eq_water_recovery = Constraint(
        expr=sum(blk.feed.properties[0].flow_mass_phase_comp["Liq", :])
        * blk.water_recovery
        / 100
        == blk.product.properties[0].flow_mass_phase_comp["Liq", "H2O"]
    )
    iscale.set_scaling_factor(blk.water_recovery, 1 / 100)
    iscale.constraint_scaling_transform(blk.eq_water_recovery, 1 / 100)

    blk.disposal_mass_flow_feed_basis = Var(
        list(blk.feed.properties[0].flow_mass_phase_comp.keys()),
        units=pyunits.kg / pyunits.m**3,
        bounds=(None, None),
    )

    @blk.Constraint(list(blk.feed.properties[0].flow_mass_phase_comp.keys()))
    def eq_disposal_mass_flow_feed_basis(blk, liq, ion):
        return (
            blk.disposal_mass_flow_feed_basis[liq, ion]
            == blk.disposal.properties[0].flow_mass_phase_comp[liq, ion]
            / blk.feed.properties[0].flow_vol_phase["Liq"]
        )


def initialize(m, solver=None, **kwargs):
    if solver is None:
        solver = get_solver()
    m.fs.feed.initialize(optarg=solver.options)
    # loop over all the process and their arcs and initialize
    # right now only supporting linear trains with no recycle
    for i, proc_dict in enumerate(m.fs.process_order):
        propagate_state(m.arc_order[i])
        if proc_dict.get("init_args") is not None:
            proc_dict["init"](
                m, proc_dict["process_block"], solver, **proc_dict.get("init_args")
            )
        else:
            proc_dict["init"](m, proc_dict["process_block"], solver)
    for arg in m.end_point_order:
        propagate_state(arg)

    # Only init if it was constructed
    if m.fs.find_component("product_mixer") is not None:
        m.fs.product_mixer.initialize(optarg=solver.options)
    m.fs.product.initialize(optarg=solver.options)
    update_feed(m.fs, solver)
    fix_conc_feed(
        m.fs,
    )
    setup_optimization(m)
    solve(m, solver)


def build_selected_processes(m, process_name, process_type, default_kwargs=None):
    m.fs.add_component(process_name, FlowsheetBlock(dynamic=False))
    block = m.fs.find_component(process_name)
    block.process_name = process_name
    if default_kwargs == None:
        default_kwargs = {}
    if "desalter" in process_type:
        desalter.build(m, block, **default_kwargs)
        m.fs.process_order.append(
            {
                "process_type": process_type,
                "process_name": process_name,
                "process_block": block,
                "init": desalter.initialize,
                "feed": block.desalter.inlet,
                "product": block.desalter.product,
                "effluent": block.desalter.brine,
                "disposal": None,
                "post_scale_func": None,
                "fix_vars_func": None,
                "unfix_opt_vars": desalter.unfix_opt_vars,
                "opt_guess": None,
                "init_args": None,
                "display_func": desalter.display,
                "default_scale_func": None,
            }
        )
    elif "separator" in process_type:
        separator.build(m, block, **default_kwargs)
        m.fs.process_order.append(
            {
                "process_type": process_type,
                "process_name": process_name,
                "process_block": block,
                "init": separator.initialize,
                "feed": block.separator.inlet,
                "product": None,
                "effluent": block.separator.treated,
                "disposal": block.separator.product,
                "post_scale_func": None,
                "fix_vars_func": None,
                "unfix_opt_vars": None,
                "opt_guess": None,
                "init_args": None,
                "display_func": separator.display,
                "default_scale_func": None,
            }
        )
    elif "valorizer" in process_type:
        default_kwargs.update({"valorizer_costing": True})
        separator.build(m, block, **default_kwargs)
        m.fs.process_order.append(
            {
                "process_type": process_type,
                "process_name": process_name,
                "process_block": block,
                "init": separator.initialize,
                "feed": block.separator.inlet,
                "product": None,
                "effluent": block.separator.treated,
                "disposal": None,
                "post_scale_func": None,
                "fix_vars_func": None,
                "unfix_opt_vars": None,
                "opt_guess": None,
                "init_args": None,
                "display_func": separator.display,
                "default_scale_func": None,
            }
        )
    else:
        raise TypeError("Selected process type {} unavailable".format(process_type))


def setup_optimization(m):
    """unfix all vars in processes for optimization"""
    for i, process_dict in enumerate(m.fs.process_order):
        if process_dict.get("unfix_opt_vars") != None:
            process_dict["unfix_opt_vars"](m, process_dict["process_block"])
    """ enable objective function """
    # m.fs.water_recovery.fix(95)
    m.fs.cost_objective.activate()


def solve(m, solver=None, tee=None):
    if m.find_component("fs") is None:
        update_feed(m, solver)
    else:
        update_feed(m.fs, solver)
    _logger.info(f"DOFS:{degrees_of_freedom(m)}")
    if solver is None:
        solver = get_solver()

    result = solver.solve(m)
    if m.find_component("fs") is None:
        fix_conc_feed(m)
        display_processes(m)
    else:
        fix_conc_feed(m.fs)
        display_processes(m.fs)
    return result


def fix_conc_feed(blk):
    blk.feed.properties[0].flow_mass_phase_comp.unfix()
    blk.feed.properties[0].conc_mass_phase_comp.fix()
    blk.feed.properties[0].conc_mass_phase_comp["Liq", "H2O"].unfix()
    blk.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"].fix()


def update_feed(blk, solver=None):
    """used to update feed composition based on what user enters
    in ui, which uses concentration, rather then mass flow basis"""
    if solver == None:
        solver = get_solver()
    _logger.info("solved feed")
    fix_conc_feed(blk)
    solver.solve(blk.feed.properties[0])
    blk.feed.properties[0].conc_mass_phase_comp.unfix()
    blk.feed.properties[0].flow_mass_phase_comp.fix()


def display_processes(blk):
    _logger.info("--------Display results for {}--------".format("Overall process"))
    _logger.info("LCOW {}  $/m3".format(value(blk.costing.LCOW)))
    for name, var in blk.costing.LCOW_unit.items():
        _logger.info(f"{name} LCOW {value(var)}")
    _logger.info("Water recovery {}%".format(blk.water_recovery.value))
    _logger.info(
        f"Feed flow (kg/s) {value(blk.feed.properties[0].flow_mass_phase_comp['Liq','H2O'])}"
    )
    _logger.info(
        f"Product flow (kg/s) {value(blk.product.properties[0].flow_mass_phase_comp['Liq','H2O'])}"
    )
    _logger.info(
        f"Disposal flow (kg/s) {value(blk.disposal.properties[0].flow_mass_phase_comp['Liq','H2O'])}"
    )

    _logger.info(f"Annual feed cost ($) {value(blk.feed.annual_cost)}")
    _logger.info(f"Annual product cost ($) {value(blk.product.annual_cost)}")
    _logger.info(f"Annual disposal cost ($) {value(blk.disposal.annual_cost)}")

    _logger.info("--------------------------------")
    for proc_dict in blk.process_order:
        if proc_dict.get("display_func") is not None:
            _logger.info(
                "--------Display results for {}--------".format(
                    proc_dict["process_name"]
                )
            )
            proc_dict.get("display_func")(blk, proc_dict["process_block"])
            _logger.info("--------------------------------")


if __name__ == "__main__":
    main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

###############################################################################
#
# OLI Systems, Inc. Copyright © 2022, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of OLI Systems, Inc. nor the names of any contributors to
# the software made available herein may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the
# features, functionality or performance of the source code ("Enhancements") to anyone; however,
# if you choose to make your Enhancements available either publicly, or directly to OLI Systems, Inc.,
# without imposing a separate written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license to install, use, modify, prepare
# derivative works, incorporate into other computer software, distribute, and sublicense such enhancements
# or derivative works thereof, in binary and source code form.
###############################################################################

__author__ = "Adam Atia, Adi Bannady, Paul Vecchiarelli"

import logging

import sys
import json
import time
from pyomo.common.dependencies import attempt_import

requests, requests_available = attempt_import("requests", defer_import=False)
from watertap.tools.oli_api.util.watertap_to_oli_helper_functions import get_oli_name


_logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "OLIAPI - %(asctime)s - %(levelname)s - %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)


class OLIApi:
    """
    A class to wrap OLI Cloud API calls and access functions for interfacing with WaterTAP.
    """

    def __init__(self, credential_manager, interactive_mode=True, debug_level="INFO"):
        """
        Construct all necessary attributes for OLIApi class.

        :param credential_manager_class: class used to manage credentials
        :param interactive_mode: enables direct interaction with user through prompts
        :param debug_level: string defining level of logging activity
        """

        self.credential_manager = credential_manager
        self.interactive_mode = interactive_mode
        if self.interactive_mode:
            _logger.info(
                "OLI running in interactive mode, disable with 'interactive_mode=False'"
            )
        if debug_level == "INFO":
            _logger.setLevel(logging.INFO)
        else:
            _logger.setLevel(logging.DEBUG)

        # TODO: unknown bug where only "liquid1" phase is found in Flash analysis
        self.valid_phases = ["liquid1", "vapor", "solid", "liquid2"]

    # binds OLIApi instance to context manager
    def __enter__(self):
        self.session_dbs_files = []
        return self

    # return False if no exceptions raised
    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        # delete all .dbs files created during session
        _logger.info(
            f"Exiting: deleting {len(self.session_dbs_files)} remaining DBS files created during the session that were not marked by keep_file=True."
        )
        self.dbs_file_cleanup(self.session_dbs_files)
        return False

    def _prompt(self, msg, default=""):
        if self.interactive_mode:
            msg = msg + "Enter [y]/n to proceed."
            return input(msg)
        else:
            msg = msg + "To choose [y]/n to this action, set interactive_mode=True."
            _logger.info(msg)
            return default

    def upload_dbs_file(self, dbs_file_path, keep_file=False):
        """
        Upload a DBS file to OLI Cloud given a full file path.

        :param dbs_file_path: string path to DBS file
        :param keep_file: bool to remove (default) or ignore DBS file during session cleanup

        :return dbs_file_id: string name for DBS file ID
        """

        with open(dbs_file_path, "rb") as file:
            req = requests.post(
                self.credential_manager.upload_dbs_url,
                headers=self.credential_manager.headers,
                files={"files": file},
            )
        dbs_file_id = _request_status_test(req, ["UPLOADED"])["file"][0]["id"]
        if bool(dbs_file_id):
            if not keep_file:
                self.session_dbs_files.append(dbs_file_id)
            _logger.info(f"Uploaded DBS file ID is {dbs_file_id}")
            return dbs_file_id
        else:
            raise RuntimeError("Unexpected failure getting DBS file ID.")

    def generate_dbs_file(
        self,
        inflows,
        thermo_framework=None,
        model_name=None,
        phases=None,
        databanks=None,
        keep_file=False,
    ):
        """
        Generate a DBS file on OLI Cloud given chemistry inputs.

        :param inflows: dictionary with inflows and optional custom parameters
        :param thermo_framework: string name of thermodynamic databank to use
        :param model_name: string name of model OLI will use
        :param phases: container dict for chemistry model parameters
        :param databanks: list of databanks to include in DBS file
        :param keep_file: bool to remove (default) or ignore DBS file during session cleanup

        :return dbs_file_id: string name for DBS file ID
        """

        dbs_file_inputs = {
            "thermodynamicFramework": None,
            "modelName": None,
            "phases": None,
            "inflows": None,
            "privateDatabanks": None,
        }

        if not inflows:
            raise RuntimeError("Inflows must be defined for Water Analysis.")
        solute_list = [{"name": get_oli_name(k)} for k in inflows]
        if not solute_list:
            raise RuntimeError("No inflows extracted from {inflows}.")
        dbs_file_inputs["inflows"] = solute_list

        if thermo_framework is not None:
            if thermo_framework in ["MSE (H3O+ ion)", "Aqueous (H+ ion)"]:
                dbs_file_inputs["thermodynamicFramework"] = thermo_framework
            else:
                raise RuntimeError(
                    "Failed DBS file generation. "
                    + f"Unexpected thermo_framework: {thermo_framework}."
                )
        else:
            dbs_file_inputs["thermodynamicFramework"] = "MSE (H3O+ ion)"

        if model_name is not None:
            dbs_file_inputs["modelName"] = str(model_name)
        else:
            dbs_file_inputs["modelName"] = "OLI_analysis"

        if phases is not None:
            invalid_phases = [p for p in phases if p not in self.valid_phases]
            if invalid_phases:
                raise RuntimeError(
                    "Failed DBS file generation. "
                    + f"Unexpected phase(s): {invalid_phases}"
                )
            dbs_file_inputs["phases"] = phases
        else:
            dbs_file_inputs["phases"] = ["liquid1", "solid", "vapor"]

        valid_databanks = ["XSC"]
        if databanks is not None:
            invalid_databanks = [db for db in databanks if db not in valid_databanks]
            if invalid_databanks:
                raise RuntimeError(
                    "Failed DBS file generation. "
                    + f"Unexpected databanks(s): {invalid_databanks}"
                )
            dbs_file_inputs["privateDatabanks"] = databanks

        dbs_dict = {
            "method": "chemistrybuilder.generateDBS",
            "params": {k: v for k, v in dbs_file_inputs.items() if v is not None},
        }
        _logger.debug(f"DBS input dictionary: {dbs_dict}")
        req = requests.post(
            self.credential_manager.dbs_url,
            headers=self.credential_manager.update_headers(
                {"Content-Type": "application/json"}
            ),
            data=json.dumps(dbs_dict),
        )
        dbs_file_id = _request_status_test(req, ["SUCCESS"])["data"]["id"]
        if bool(dbs_file_id):
            if not keep_file:
                self.session_dbs_files.append(dbs_file_id)
            _logger.info(f"Generated DBS file ID is {dbs_file_id}")
            return dbs_file_id
        else:
            raise RuntimeError("Unexpected failure getting DBS file ID.")

    def get_dbs_file_summary(self, dbs_file_id):
        """
        Get chemistry info and flash history for a DBS file.

        :param dbs_file_id: string name for DBS file ID

        :return dbs_file_summary: dictionary containing JSON results from OLI Cloud
        """

        _logger.info(f"Getting summary for {dbs_file_id} ...")
        chemistry_info = self.call("chemistry-info", dbs_file_id)
        req_flash_hist = requests.get(
            f"{self.credential_manager.engine_url}/flash/history/{dbs_file_id}",
            headers=self.credential_manager.headers,
        )
        flash_history = _request_status_test(req_flash_hist, None)["data"]
        dbs_file_summary = {
            "chemistry_info": chemistry_info,
            "flash_history": flash_history,
        }
        missing_keys = [k for k, v in dbs_file_summary.items() if not v]
        if len(missing_keys) > 0:
            raise RuntimeError(
                " Failed to create DBS file summary. " + f"Missing {missing_keys}."
            )
        _logger.info(f"Completed DBS file summarization")
        return dbs_file_summary

    def get_user_dbs_file_ids(self):
        """
        Get all DBS files on user's cloud.

        :return user_dbs_file_ids: list of user DBS files saved on OLI Cloud
        """

        _logger.info("Getting DBS file IDs for user ...")
        req = requests.get(
            self.credential_manager.dbs_url,
            headers=self.credential_manager.headers,
        )
        user_dbs_file_ids = [
            k["fileId"] for k in _request_status_test(req, None)["data"]
        ]
        _logger.info(f"{len(user_dbs_file_ids)} DBS files found for user")
        return user_dbs_file_ids

    def dbs_file_cleanup(self, dbs_file_ids=None):
        """
        Delete all (or specified) DBS files on OLI Cloud.

        :param dbs_file_ids: list of DBS file IDs to delete
        """

        if dbs_file_ids is None:
            _logger.info(
                "No DBS file IDs were provided to the dbs_file_cleanup method. Checking user's cloud account for DBS file IDs."
            )
            dbs_file_ids = self.get_user_dbs_file_ids()
            if not len(dbs_file_ids):
                _logger.info("No DBS file IDs were found on the user's cloud account.")
                return

        r = self._prompt(f"WaterTAP will delete {len(dbs_file_ids)} DBS files. ", "y")
        if (r.lower() == "y") or (r == ""):
            for dbs_file_id in dbs_file_ids:
                _logger.info(f"Deleting {dbs_file_id} ...")
                req = requests.request(
                    "DELETE",
                    f"{self.credential_manager._delete_dbs_url}{dbs_file_id}",
                    headers=self.credential_manager.headers,
                )
                req = _request_status_test(req, ["SUCCESS"])

                if req["status"] == "SUCCESS":
                    # Remove the file from session_dbs_files list if it is there, otherwise an error will occur upon exit when this method is called again and already deleted files will remain on the list for deletion. Thus, an error can occur if there is no existing ID to delete.
                    if dbs_file_id in self.session_dbs_files:
                        self.session_dbs_files.remove(dbs_file_id)
                        _logger.info(
                            f"File {dbs_file_id} deleted and removed from session_dbs_files list."
                        )
                    else:
                        _logger.info(f"File {dbs_file_id} deleted.")

    def get_corrosion_contact_surfaces(self, dbs_file_id):
        """
        Get list of valid contact surfaces for corrosion analysis, given a DBS file ID.

        :param dbs_file_id: string name for DBS file ID

        :return contact_surfaces: JSON results from OLI Cloud
        """

        _logger.info(f"Checking chemistry info for {dbs_file_id} ...")
        chemistry_info = self.call("chemistry-info", dbs_file_id)
        framework = chemistry_info["result"]["thermodynamicFramework"]
        if framework != "AQ":
            raise RuntimeError(
                f"Invalid thermodynamic framework for {dbs_file_id}. "
                f"Corrosion analyzer requires 'Aqueous (H+ ion)', not {framework}."
            )

        _logger.info(f"Getting contact surfaces for {dbs_file_id} ...")
        contact_surfaces = self.call("corrosion-contact-surface", dbs_file_id)
        return contact_surfaces

    def _get_flash_mode(self, dbs_file_id, flash_method, burst_job_tag=None):

        if not bool(flash_method):
            raise IOError("Specify a flash method to run.")
        if not bool(dbs_file_id):
            raise IOError("Specify a DBS file ID to flash.")

        headers = self.credential_manager.headers
        base_url = self.credential_manager.engine_url
        valid_get_flashes = ["corrosion-contact-surface", "chemistry-info"]
        valid_post_flashes = [
            "isothermal",
            "corrosion-rates",
            "wateranalysis",
            "bubblepoint",
        ]

        if flash_method in [
            "bubblepoint",
            # TODO: uncomment the methods below only after trying and testing
            # "dewpoint",
            # "vapor-amount",
            # "vapor-fraction",
            # "isochoric",
        ]:
            dbs_summary = self.get_dbs_file_summary(dbs_file_id)
            phase_list = dbs_summary["chemistry_info"]["result"]["phases"]

            if "vapor" not in phase_list:
                raise RuntimeError(
                    "A vapor function ('{flash_method}') was called without included 'vapor' as a phase in the model"
                )

        if flash_method in valid_get_flashes:
            mode = "GET"
            url = f"{base_url}/file/{dbs_file_id}/{flash_method}"

        elif flash_method in valid_post_flashes:
            mode = "POST"
            url = f"{base_url}/flash/{dbs_file_id}/{flash_method}"
            headers = self.credential_manager.update_headers(
                {"Content-Type": "application/json"},
            )
            if burst_job_tag is not None:
                url = "{}?{}".format(
                    url, "burst={}_{}".format("watertap_burst", burst_job_tag)
                )
        else:
            valid_flashes = [*valid_get_flashes, *valid_post_flashes]
            raise RuntimeError(
                f" Unexpected value for flash_method: {flash_method}. "
                + "Valid values: {', '.join(valid_flashes)}."
            )
        return mode, url, headers

    def process_request_list(self, requests, **kwargs):
        """
        Process a list of flash calculation requests for OLI.

        :param requests: list of request dictionaries containing flash_method, dbs_file_id, and json_input
        """

        _logger.info("Collecting requested samples in serial mode ...")
        result_list = []
        acquire_timer = time.time()
        for idx, request in enumerate(requests):
            _logger.info(f"Submitting sample #{idx+1} of {len(requests)} ...")
            result = self.call(**request)
            result["submitted_requests"] = request
            result_list.append(result)
        acquire_time = time.time() - acquire_timer
        num_samples = len(requests)
        _logger.info(
            f"Finished all {num_samples} jobs from OLI. "
            + f"Total: {acquire_time} s, "
            + f"Rate: {acquire_time/num_samples} s/sample"
        )
        return result_list

    def call(
        self,
        flash_method=None,
        dbs_file_id=None,
        input_params=None,
        poll_time=0.5,
        max_request=100,
        **kwargs,
    ):
        """
        Make a call to the OLI Cloud API.

        :param flash_method: string indicating flash method
        :param dbs_file_id: string indicating DBS file
        :param input_params: dictionary for flash calculation inputs
        :param poll_time: seconds between each poll
        :param max_request: maximum number of times to try request before failure

        :return result: dictionary for JSON output result
        """

        mode, url, headers = self._get_flash_mode(dbs_file_id, flash_method)
        try:
            req = requests.request(
                mode, url, headers=headers, data=json.dumps(input_params)
            )
            req_json = _request_status_test(req, ["SUCCESS"])
        except requests.JSONDecodeError:
            delay = 1
            time.sleep(delay)
            _logger.debug(
                f"JSONDecodeError occurred. Trying to convert response object to JSON again."
            )
            req = requests.request(
                mode, url, headers=headers, data=json.dumps(input_params)
            )
            req_json = _request_status_test(req, ["SUCCESS"])
        result_link = _get_result_link(req_json)
        result = _poll_result_link(result_link, headers, max_request, poll_time)
        return result


def _get_result_link(req_json):
    """
    Get result link from OLI Cloud request.

    :param req_json: JSON containing result of request

    :return result_link: string indicating URL to access call results
    """

    if "data" in req_json:
        if "status" in req_json["data"]:
            if req_json["data"]["status"] in ["IN QUEUE", "IN PROGRESS"]:
                if "resultsLink" in req_json["data"]:
                    result_link = req_json["data"]["resultsLink"]
            if "resultsLink" in req_json["data"]:
                result_link = req_json["data"]["resultsLink"]
    if not result_link:
        raise RuntimeError(f"Failed to get 'resultsLink'. Response: {req_json.json()}")
    return result_link


def _request_status_test(req, target_keys):
    """
    Check result of OLI request (except async).

    :param req: response object
    :param target_keys: list containing key(s) that indicate successful request

    :return req_json: response object converted to JSON
    """
    req_json = req.json()

    func_name = sys._getframe().f_back.f_code.co_name
    _logger.debug(f"{func_name} response: {req_json}")

    if req.status_code == 200:
        if target_keys:
            if "status" in req_json:
                if req_json["status"] in target_keys:
                    return req_json
        else:
            return req_json
    raise RuntimeError(
        f"Failure in {func_name}. Response: {req_json}. Status Code: {req.status_code}."
    )


def _poll_result_link(result_link, headers, max_request, poll_time):
    """
    Poll result link from OLI Flash calculation request.

    :param result_link: string indicating URL to access call results
    :param headers: dictionary for OLI Cloud headers
    :param max_request: integer for number of requests before poll limit error
    :param poll_time: float for time in between poll requests

    return result: JSON containing results from successful Flash calculation
    """

    for _ in range(max_request):
        time.sleep(poll_time)
        result_req = requests.get(result_link, headers=headers)
        result_req = _request_status_test(
            result_req, ["IN QUEUE", "IN PROGRESS", "PROCESSED", "FAILED"]
        )
        _logger.info(f"Polling result link: {result_req['status']}")
        if result_req["status"] in ["PROCESSED", "FAILED"]:
            if result_req["data"]:
                result = result_req["data"]
                return result
    raise RuntimeError("Poll limit exceeded.")



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
    Block,
)
from pyomo.network import Port

from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from watertap.unit_models.crystallizer import Crystallization
from watertap.flowsheets.crystallization.sim_simple_crystallizer import main as cryst_ex
from watertap.property_models.unit_specific.cryst_prop_pack import NaClParameterBlock
from watertap.costing import CrystallizerCostType


class TestCrystallizerBuild:
    @pytest.fixture(scope="class")
    def crystallizer(self):
        m = cryst_ex()
        return m

    @pytest.mark.unit
    def test_model_properties(self, crystallizer):
        m = crystallizer

        # Test basic build
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties, NaClParameterBlock)
        assert isinstance(m.fs.costing, Block)
        assert isinstance(m.fs.crystallizer, Crystallization)

        # Test port
        assert isinstance(m.fs.crystallizer.inlet, Port)
        assert isinstance(m.fs.crystallizer.vapor, Port)
        assert isinstance(m.fs.crystallizer.solids, Port)
        assert isinstance(m.fs.crystallizer.outlet, Port)

        # Test consting
        assert isinstance(m.fs.crystallizer.costing, Block)
        assert isinstance(m.fs.crystallizer.costing.capital_cost, Var)

        var_str_list = [
            "total_capital_cost",
            "total_operating_cost",
        ]
        for var_str in var_str_list:
            var = getattr(m.fs.costing, var_str)
            assert isinstance(var, Var)

        # Test the crystallizer properties
        # test configrations
        assert len(m.fs.crystallizer.config) == 4
        assert not m.fs.crystallizer.config.dynamic
        assert not m.fs.crystallizer.config.has_holdup
        assert m.fs.crystallizer.config.property_package is m.fs.properties
        assert "H2O" in m.fs.properties.component_list
        assert "NaCl" in m.fs.properties.component_list
        assert (
            m.fs.crystallizer.costing.config["costing_method_arguments"]["cost_type"]
            == CrystallizerCostType.mass_basis
        )

    # @pytest.mark.component
    def test_final_simulation_conditions(self, crystallizer):
        m = crystallizer
        assert degrees_of_freedom(m) == 0
        assert value(m.fs.crystallizer.pressure_operating) == pytest.approx(
            11992, rel=1e-3
        )
        assert value(
            m.fs.crystallizer.product_volumetric_solids_fraction
        ) == pytest.approx(0.0874, rel=1e-3)
        assert value(m.fs.crystallizer.dens_mass_magma) == pytest.approx(
            184.93, rel=1e-3
        )
        assert value(
            m.fs.crystallizer.vapor.flow_mass_phase_comp[0.0, "Vap", "H2O"]
        ) == pytest.approx(21.505, rel=1e-3)
        assert value(
            m.fs.crystallizer.solids.flow_mass_phase_comp[0.0, "Sol", "NaCl"]
        ) == pytest.approx(4.0815, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
)

from idaes.core import (
    FlowsheetBlock,
    UnitModelCostingBlock,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import initialization_tester

from watertap.costing import WaterTAPCosting
from watertap.core.util.initialization import check_dof
from watertap.core.solvers import get_solver
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock
from watertap.unit_models.ion_exchange_0D import IonExchange0D
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

__author__ = "Kurban Sitterley"

solver = get_solver()
zero = 1e-8
relative_tolerance = 1e-3


def build_langmuir():

    target_ion = "Ca_2+"
    ion_props = {
        "solute_list": [target_ion],
        "diffusivity_data": {("Liq", target_ion): 9.2e-10},
        "mw_data": {"H2O": 0.018, target_ion: 0.04},
        "charge": {target_ion: 2},
    }
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(**ion_props)
    ix_config = {
        "property_package": m.fs.properties,
        "target_ion": target_ion,
    }
    m.fs.unit = ix = IonExchange0D(**ix_config)

    ix.process_flow.properties_in.calculate_state(
        var_args={
            ("flow_vol_phase", "Liq"): 0.5,
            ("conc_mass_phase_comp", ("Liq", target_ion)): 0.1,
            ("pressure", None): 101325,
            ("temperature", None): 298,
        },
        hold_state=True,
    )

    ix.process_flow.properties_in[0].flow_mass_phase_comp[...]
    ix.process_flow.properties_out[0].flow_mass_phase_comp[...]
    ix.regeneration_stream[0].flow_mass_phase_comp[...]

    ix.service_flow_rate.fix(15)
    ix.langmuir[target_ion].fix(0.9)
    ix.resin_max_capacity.fix(3)
    ix.bed_depth.fix(1.7)
    ix.dimensionless_time.fix()
    ix.number_columns.fix(8)
    ix.resin_diam.fix()
    ix.resin_bulk_dens.fix()
    ix.bed_porosity.fix()

    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e-4, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 10, index=("Liq", "Ca_2+")
    )

    iscale.calculate_scaling_factors(m)

    return m


class TestIXLangmuir(UnitTestHarness):
    def configure(self):
        m = build_langmuir()

        self.default_zero = zero
        self.default_relative_tolerance = relative_tolerance

        self.unit_solutions[m.fs.unit.c_norm["Ca_2+"]] = 0.4919
        self.unit_solutions[m.fs.unit.col_height] = 3.488
        self.unit_solutions[m.fs.unit.col_diam] = 3.351
        self.unit_solutions[m.fs.unit.col_height_to_diam_ratio] = 1.0408
        self.unit_solutions[m.fs.unit.number_columns] = 8
        self.unit_solutions[m.fs.unit.t_breakthru] = 52360
        self.unit_solutions[m.fs.unit.vel_bed] = 0.007083
        self.unit_solutions[m.fs.unit.N_Re] = 4.958
        self.unit_solutions[m.fs.unit.N_Sc["Ca_2+"]] = 1086
        self.unit_solutions[m.fs.unit.N_Sh["Ca_2+"]] = 26.296
        self.unit_solutions[m.fs.unit.N_Pe_particle] = 0.10782
        self.unit_solutions[m.fs.unit.N_Pe_bed] = 261.8
        self.unit_solutions[m.fs.unit.resin_eq_capacity] = 1.554
        self.unit_solutions[m.fs.unit.resin_unused_capacity] = 1.445
        self.unit_solutions[m.fs.unit.langmuir["Ca_2+"]] = 0.9
        self.unit_solutions[m.fs.unit.mass_removed["Ca_2+"]] = 65300
        self.unit_solutions[m.fs.unit.num_transfer_units] = 35.5483
        self.unit_solutions[m.fs.unit.dimensionless_time] = 1
        self.unit_solutions[m.fs.unit.partition_ratio] = 217.669
        self.unit_solutions[m.fs.unit.fluid_mass_transfer_coeff["Ca_2+"]] = 3.4561e-05
        self.unit_solutions[m.fs.unit.bw_flow] = 0.09803
        self.unit_solutions[m.fs.unit.bed_expansion_frac] = 0.46395
        self.unit_solutions[m.fs.unit.col_vol_tot] = 246.26
        self.unit_solutions[m.fs.unit.lh] = 0.0
        self.unit_solutions[m.fs.unit.separation_factor["Ca_2+"]] = 1.111
        self.unit_solutions[m.fs.unit.rate_coeff["Ca_2+"]] = 0.00021159
        self.unit_solutions[m.fs.unit.HTU["Ca_2+"]] = 0.04782
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 499.95
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        ] = 0.05
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 499.95
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        ] = 1.1458e-4
        self.unit_solutions[
            m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp["Liq", "Ca_2+"]
        ] = 0.049885
        self.unit_solutions[
            m.fs.unit.process_flow.mass_transfer_term[0.0, "Liq", "Ca_2+"]
        ] = -1.24713

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Ca_2+"
                ],
                "out": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Ca_2+"
                ]
                + m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp[
                    "Liq", "Ca_2+"
                ],
            },
        }

        return m

    @pytest.mark.component
    def test_costing(self):
        m = build_langmuir()
        ix = m.fs.unit

        m.fs.costing = WaterTAPCosting()
        ix.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ix.process_flow.properties_out[0].flow_vol_phase["Liq"])
        m.fs.costing.add_specific_energy_consumption(
            ix.process_flow.properties_out[0].flow_vol_phase["Liq"], name="SEC"
        )

        check_dof(m, fail_flag=True)
        initialization_tester(m)
        results = solver.solve(m, tee=True)
        assert_optimal_termination(results)

        sys_cost_results = {
            "aggregate_capital_cost": 3993072.4,
            "aggregate_fixed_operating_cost": 36893.3,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 39.498,
            "aggregate_flow_NaCl": 22838957.969,
            "aggregate_flow_costs": {
                "electricity": 24237.08,
                "NaCl": 2079295.202,
            },
            "total_capital_cost": 3993072.4698,
            "total_operating_cost": 2049864.542,
            "LCOW": 0.1724829,
            "SEC": 0.021945,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)

        ix_cost_results = {
            "capital_cost": 3993072.4,
            "fixed_operating_cost": 36893.3,
            "capital_cost_vessel": 101131.8,
            "capital_cost_resin": 81985.1,
            "capital_cost_regen_tank": 215778.2,
            "capital_cost_backwash_tank": 132704.7,
            "operating_cost_hazardous": 0,
            "flow_mass_regen_soln": 22838957.9,
            "total_pumping_power": 39.4985,
            "backwash_tank_vol": 174042.7,
            "regeneration_tank_vol": 79251.6,
            "direct_capital_cost": 1996536.2,
        }

        for v, r in ix_cost_results.items():
            mv = getattr(m.fs.unit.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)


def build_freundlich():

    target_ion = "Cl_-"

    ion_props = {
        "solute_list": [target_ion],
        "diffusivity_data": {("Liq", target_ion): 1e-9},
        "mw_data": {"H2O": 0.018, target_ion: 35.45e-3},
        "charge": {target_ion: -1},
    }

    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(**ion_props)
    ix_config = {
        "property_package": m.fs.properties,
        "target_ion": target_ion,
        "isotherm": "freundlich",
        "regenerant": "NaOH",
        "hazardous_waste": True,
    }
    m.fs.unit = ix = IonExchange0D(**ix_config)

    ix.process_flow.properties_in.calculate_state(
        var_args={
            ("flow_vol_phase", "Liq"): 0.5,
            ("conc_mass_phase_comp", ("Liq", target_ion)): 1e-6,
            ("pressure", None): 101325,
            ("temperature", None): 298,
        },
        hold_state=True,
    )

    ix.process_flow.properties_in[0].flow_mass_phase_comp[...]
    ix.process_flow.properties_out[0].flow_mass_phase_comp[...]
    ix.regeneration_stream[0].flow_mass_phase_comp[...]

    ix.freundlich_n.fix(1.2)
    ix.bv_50.fix(20000)
    ix.bv.fix(18000)
    ix.resin_bulk_dens.fix(0.72)
    ix.bed_porosity.fix()
    ix.vel_bed.fix(6.15e-3)
    ix.resin_diam.fix(6.75e-4)
    ix.number_columns.fix(16)
    ix.service_flow_rate.fix(15)
    ix.c_norm.fix(0.25)

    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e-4, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e6, index=("Liq", "Cl_-")
    )
    iscale.calculate_scaling_factors(m)

    return m


class TestIXFreundlich(UnitTestHarness):
    def configure(self):
        m = build_freundlich()

        self.default_zero = zero
        self.default_relative_tolerance = relative_tolerance
        self.unit_solutions[m.fs.unit.col_height] = 3.160
        self.unit_solutions[m.fs.unit.col_diam] = 2.5435
        self.unit_solutions[m.fs.unit.N_Sh["Cl_-"]] = 24.083
        self.unit_solutions[m.fs.unit.N_Pe_particle] = 0.09901
        self.unit_solutions[m.fs.unit.N_Pe_bed] = 216.51
        self.unit_solutions[m.fs.unit.tb_traps[0]] = 0
        self.unit_solutions[m.fs.unit.tb_traps[1]] = 3344557
        self.unit_solutions[m.fs.unit.tb_traps[2]] = 3825939
        self.unit_solutions[m.fs.unit.tb_traps[3]] = 4034117
        self.unit_solutions[m.fs.unit.tb_traps[4]] = 4188551
        self.unit_solutions[m.fs.unit.tb_traps[5]] = 4320000
        self.unit_solutions[m.fs.unit.traps[1]] = 0.0038710157
        self.unit_solutions[m.fs.unit.traps[2]] = 0.0044572367
        self.unit_solutions[m.fs.unit.traps[3]] = 0.0048189406
        self.unit_solutions[m.fs.unit.traps[4]] = 0.0057197676
        self.unit_solutions[m.fs.unit.traps[5]] = 0.0066941563
        self.unit_solutions[m.fs.unit.c_norm_avg["Cl_-"]] = 0.02556
        self.unit_solutions[m.fs.unit.mass_transfer_coeff] = 0.159346
        self.unit_solutions[
            m.fs.unit.process_flow.mass_transfer_term[0.0, "Liq", "Cl_-"]
        ] = -1.3744e-05
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 500
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 500
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                "Liq", "Cl_-"
            ]
        ] = 5e-07
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "Cl_-"
            ]
        ] = 1.278e-08
        self.unit_solutions[
            m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp["Liq", "Cl_-"]
        ] = 4.872e-07

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ],
                "out": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ]
                + m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ],
            },
        }

        return m

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solution(self):
        m = build_freundlich()
        ix = m.fs.unit

        m.fs.costing = WaterTAPCosting()
        ix.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ix.process_flow.properties_out[0].flow_vol_phase["Liq"])
        m.fs.costing.add_specific_energy_consumption(
            ix.process_flow.properties_out[0].flow_vol_phase["Liq"], name="SEC"
        )

        check_dof(m, fail_flag=True)
        initialization_tester(m)
        results = solver.solve(m, tee=True)
        assert_optimal_termination(results)

        sys_cost_results = {
            "aggregate_capital_cost": 5116213.693,
            "aggregate_fixed_operating_cost": 323306.067,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 30.805,
            "aggregate_flow_NaOH": 279183.597,
            "aggregate_flow_costs": {
                "electricity": 18902.5053,
                "NaOH": 555415.521,
            },
            "total_capital_cost": 5116213.693,
            "total_operating_cost": 993678.703,
            "aggregate_direct_capital_cost": 2558106.846,
            "maintenance_labor_chemical_operating_cost": 153486.410,
            "total_fixed_operating_cost": 476792.478,
            "total_variable_operating_cost": 516886.223,
            "total_annualized_cost": 1505300.072,
            "LCOW": 0.1060001,
            "SEC": 0.017113,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)

        ix_cost_results = {
            "capital_cost": 5116213.6933,
            "fixed_operating_cost": 323306.0679,
            "capital_cost_vessel": 75000.3204,
            "capital_cost_resin": 54924.687,
            "capital_cost_regen_tank": 215778.261,
            "capital_cost_backwash_tank": 133603.4529,
            "operating_cost_hazardous": 276620.0837,
            "flow_mass_regen_soln": 279183.597,
            "total_pumping_power": 30.8049,
            "backwash_tank_vol": 176401.0669,
            "regeneration_tank_vol": 79251.615,
            "direct_capital_cost": 2558106.846,
        }

        for v, r in ix_cost_results.items():
            mv = getattr(m.fs.unit.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)


def build_inert():
    target_ion = "Cl_-"
    inert = "Ca_2+"

    ion_props = {
        "solute_list": [target_ion, inert],
        "diffusivity_data": {("Liq", target_ion): 1e-9, ("Liq", inert): 9.2e-10},
        "mw_data": {"H2O": 0.018, target_ion: 35.45e-3, inert: 40e-3},
        "charge": {target_ion: -1, inert: 2},
    }

    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = MCASParameterBlock(**ion_props)
    ix_config = {
        "property_package": m.fs.properties,
        "target_ion": target_ion,
        "regenerant": "single_use",
        "isotherm": "freundlich",
    }
    m.fs.unit = ix = IonExchange0D(**ix_config)

    ix.process_flow.properties_in.calculate_state(
        var_args={
            ("flow_vol_phase", "Liq"): 0.5,
            ("conc_mass_phase_comp", ("Liq", target_ion)): 1e-6,
            ("conc_mass_phase_comp", ("Liq", inert)): 1e-7,
            ("pressure", None): 101325,
            ("temperature", None): 298,
        },
        hold_state=True,
    )

    ix.process_flow.properties_in[0].flow_mass_phase_comp[...]
    ix.process_flow.properties_out[0].flow_mass_phase_comp[...]
    ix.regeneration_stream[0].flow_mass_phase_comp[...]

    ix.freundlich_n.fix(1.2)
    ix.bv_50.fix(20000)
    ix.bv.fix(18000)
    ix.resin_bulk_dens.fix(0.72)
    ix.bed_porosity.fix()
    ix.vel_bed.fix(6.15e-3)
    ix.resin_diam.fix(6.75e-4)
    ix.number_columns.fix(16)
    ix.service_flow_rate.fix(15)
    ix.c_norm.fix(0.25)
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e-4, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e5, index=("Liq", "Cl_-")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 1e5, index=("Liq", "Ca_2+")
    )
    iscale.calculate_scaling_factors(m)

    return m


class TestIXInert(UnitTestHarness):
    def configure(self):
        m = build_inert()

        self.default_zero = zero
        self.default_relative_tolerance = relative_tolerance

        self.unit_solutions[m.fs.unit.number_columns] = 16
        self.unit_solutions[m.fs.unit.col_height] = 3.160
        self.unit_solutions[m.fs.unit.col_diam] = 2.5435
        self.unit_solutions[m.fs.unit.N_Sh["Cl_-"]] = 24.083
        self.unit_solutions[m.fs.unit.N_Pe_particle] = 0.09901
        self.unit_solutions[m.fs.unit.N_Pe_bed] = 216.51
        self.unit_solutions[m.fs.unit.tb_traps[1]] = 3344557
        self.unit_solutions[m.fs.unit.tb_traps[2]] = 3825939
        self.unit_solutions[m.fs.unit.tb_traps[3]] = 4034117
        self.unit_solutions[m.fs.unit.tb_traps[4]] = 4188551
        self.unit_solutions[m.fs.unit.tb_traps[5]] = 4320000
        self.unit_solutions[m.fs.unit.c_traps[2]] = 0.07
        self.unit_solutions[m.fs.unit.c_traps[3]] = 0.13
        self.unit_solutions[m.fs.unit.c_traps[4]] = 0.19
        self.unit_solutions[m.fs.unit.c_traps[5]] = 0.25
        self.unit_solutions[m.fs.unit.traps[4]] = 0.0057197676
        self.unit_solutions[m.fs.unit.traps[5]] = 0.0066941563
        self.unit_solutions[m.fs.unit.c_norm_avg["Cl_-"]] = 0.02556
        self.unit_solutions[m.fs.unit.mass_transfer_coeff] = 0.159346
        self.unit_solutions[
            m.fs.unit.process_flow.mass_transfer_term[0.0, "Liq", "Cl_-"]
        ] = -1.37435e-05
        self.unit_solutions[
            m.fs.unit.process_flow.mass_transfer_term[0.0, "Liq", "Ca_2+"]
        ] = 0
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 500
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "H2O"
            ]
        ] = 500
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                "Liq", "Cl_-"
            ]
        ] = 5e-07
        self.unit_solutions[
            m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        ] = 5e-08
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "Cl_-"
            ]
        ] = 1.2781e-08
        self.unit_solutions[
            m.fs.unit.process_flow.properties_out[0.0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        ] = 5e-08
        self.unit_solutions[
            m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp["Liq", "Cl_-"]
        ] = 4.872e-07

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Ca_2+"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ],
                "out": m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "H2O"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Ca_2+"
                ]
                + m.fs.unit.process_flow.properties_in[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ]
                + m.fs.unit.regeneration_stream[0.0].flow_mass_phase_comp[
                    "Liq", "Cl_-"
                ],
            },
        }

        return m

    @pytest.mark.component
    def test_costing(self):

        m = build_inert()
        ix = m.fs.unit

        m.fs.costing = WaterTAPCosting()
        ix.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(ix.process_flow.properties_out[0].flow_vol_phase["Liq"])
        m.fs.costing.add_specific_energy_consumption(
            ix.process_flow.properties_out[0].flow_vol_phase["Liq"], name="SEC"
        )

        check_dof(m, fail_flag=True)
        initialization_tester(m)
        results = solver.solve(m, tee=True)
        assert_optimal_termination(results)

        sys_cost_results = {
            "aggregate_capital_cost": 4684657.16,
            "aggregate_fixed_operating_cost": 6419597.45,
            "aggregate_variable_operating_cost": 0.0,
            "aggregate_flow_electricity": 30.813,
            "aggregate_flow_costs": {"electricity": 18907.75},
            "total_capital_cost": 4684657.16,
            "total_operating_cost": 6577154.15,
            "aggregate_direct_capital_cost": 2342328.58,
            "maintenance_labor_chemical_operating_cost": 140539.72,
            "total_fixed_operating_cost": 6560137.16,
            "total_variable_operating_cost": 17016.98,
            "total_annualized_cost": 7045619.86,
            "LCOW": 0.49613827,
            "SEC": 0.017118,
        }

        for v, r in sys_cost_results.items():
            mv = getattr(m.fs.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)

        ix_cost_results = {
            "capital_cost": 4684657.16,
            "fixed_operating_cost": 6419597.45,
            "capital_cost_vessel": 75000.32,
            "capital_cost_resin": 54924.68,
            "capital_cost_regen_tank": 0,
            "capital_cost_backwash_tank": 133603.45,
            "operating_cost_hazardous": 0,
            "flow_mass_regen_soln": 0,
            "total_pumping_power": 30.813,
            "flow_vol_resin": 876.599,
            "single_use_resin_replacement_cost": 6419597.454,
            "direct_capital_cost": 2342328.58,
        }

        for v, r in ix_cost_results.items():
            mv = getattr(m.fs.unit.costing, v)
            if mv.is_indexed():
                for i, s in r.items():
                    assert pytest.approx(s, rel=1e-3) == value(mv[i])
            else:
                assert pytest.approx(r, rel=1e-3) == value(mv)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    check_optimal_termination,
    units as pyunits,
    Var,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

from idaes.core import (
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from watertap.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog
from watertap.costing.unit_models.steam_ejector import (
    cost_steam_ejector,
)

from watertap.core import InitializationMixin

_log = idaeslog.getLogger(__name__)

__author__ = "Elmira Shamlou"


@declare_process_block_class("SteamEjector")
class SteamEjectorData(InitializationMixin, UnitModelBlockData):
    """
    Steam Ejector model for thermal vapor compression
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    def build(self):
        super().build()

        # State blocks
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = True  # motive steam inlet
        self.properties_motive_steam = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of motive steam inlet",
            **tmp_dict,
        )

        tmp_dict["defined_state"] = True  # entrained vapor inlet
        self.properties_entrained_vapor = (
            self.config.property_package.state_block_class(
                self.flowsheet().config.time,
                doc="Material properties of entrained vapor inlet",
                **tmp_dict,
            )
        )

        tmp_dict["defined_state"] = False
        self.properties_discharge_mix = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of discharge mix",
            **tmp_dict,
        )

        # Ports
        self.add_port(name="inlet_motive_steam", block=self.properties_motive_steam)
        self.add_port(
            name="inlet_entrained_vapor", block=self.properties_entrained_vapor
        )
        self.add_port(name="outlet_discharge_mix", block=self.properties_discharge_mix)

        self.PCF = Var(
            initialize=1.0,
            units=pyunits.dimensionless,
            doc="Pressure correction factor",
        )
        self.TCF = Var(
            initialize=1.0,
            bounds=(None, 4),
            units=pyunits.dimensionless,
            doc="Temperature correction factor",
        )
        self.entrainment_ratio = Var(
            initialize=0.5,
            bounds=(0, 4),
            units=pyunits.dimensionless,
            doc="Entrainment ratio",
        )

        # The bounds of compression ratio variable corresponds to the validity range of entrainment ratio semi-empirical model
        self.compression_ratio = Var(
            initialize=2,
            bounds=(1.89, None),
            units=pyunits.dimensionless,
            doc="Compression ratio",
        )

        # Empirical correlation constraints
        @self.Constraint(doc="Pressure Correction Factor (El-Dessouky, 1997)")
        def eq_PCF(b):
            A_PCF = 3e-7 / pyunits.kPa**2
            B_PCF = -0.0009 / pyunits.kPa
            C_PCF = 1.6101
            return (
                b.PCF
                == A_PCF
                * pyunits.convert(
                    b.properties_motive_steam[0].pressure, to_units=pyunits.kPa
                )
                ** 2
                - B_PCF
                * pyunits.convert(
                    b.properties_motive_steam[0].pressure, to_units=pyunits.kPa
                )
                + C_PCF
            )

        @self.Constraint(doc="Temperature Correction Factor (El-Dessouky, 1997)")
        def eq_TCF(b):
            A_TCF = 2e-8 / (pyunits.K**2)
            B_TCF = 0.0006 / pyunits.K
            C_TCF = 1.0047
            return (
                b.TCF
                == A_TCF
                * (b.properties_entrained_vapor[0].temperature - 273.15 * pyunits.K)
                ** 2
                - B_TCF
                * (b.properties_entrained_vapor[0].temperature - 273.15 * pyunits.K)
                + C_TCF
            )

        @self.Constraint(
            doc="Entrainment Ratio semi-empirical model (El-Dessouky, 1997)"
        )
        def eq_entrainment_ratio_model(b):
            # Convert pressures to kPa
            P_discharge_kPa = pyunits.convert(
                b.properties_discharge_mix[0].pressure, to_units=pyunits.kPa
            )
            P_entrained_kPa = pyunits.convert(
                b.properties_entrained_vapor[0].pressure, to_units=pyunits.kPa
            )
            P_motive_kPa = pyunits.convert(
                b.properties_motive_steam[0].pressure, to_units=pyunits.kPa
            )

            P_discharge_dim = P_discharge_kPa / (1 * pyunits.kPa)  # Dimensionless
            P_entrained_dim = P_entrained_kPa / (1 * pyunits.kPa)  # Dimensionless
            P_motive_dim = P_motive_kPa / (1 * pyunits.kPa)  # Dimensionless

            return b.entrainment_ratio * b.TCF == (
                0.296
                * (P_discharge_dim**1.19)
                / (P_entrained_dim**1.04)
                * (P_motive_dim / P_entrained_dim) ** 0.015
                * b.PCF
            )

        @self.Constraint(doc="Entrainment Ratio Definition (El-Dessouky, 1997)")
        def eq_entrainment_ratio(b):
            return (
                b.entrainment_ratio
                == b.properties_motive_steam[0].flow_mass_phase_comp["Vap", "H2O"]
                / b.properties_entrained_vapor[0].flow_mass_phase_comp["Vap", "H2O"]
            )

        @self.Constraint(doc="Compression Ratio Definition")
        def eq_compression_ratio(b):
            return (
                b.compression_ratio
                == b.properties_discharge_mix[0].pressure
                / b.properties_entrained_vapor[0].pressure
            )

        @self.Constraint(self.flowsheet().time, doc="Mass Balance")
        # assume vapor enters the ejector after passing through the demister
        def eq_mass_balance(b, t):
            lb = b.properties_discharge_mix[t].flow_mass_phase_comp["Liq", "H2O"].lb
            b.properties_discharge_mix[t].flow_mass_phase_comp["Liq", "H2O"].fix(lb)
            return b.properties_discharge_mix[t].flow_mass_phase_comp["Vap", "H2O"] == (
                b.properties_motive_steam[t].flow_mass_phase_comp["Vap", "H2O"]
                + b.properties_entrained_vapor[t].flow_mass_phase_comp["Vap", "H2O"]
            )

        # Assumes the discharge mix pressure to be equal to its saturation pressure
        @self.Constraint(
            self.flowsheet().time, doc="Discharge Temperature and Pressure Relationship"
        )
        def eq_discharge_temperature_pressure(b, t):
            return (
                b.properties_discharge_mix[t].pressure
                == b.properties_discharge_mix[t].pressure_sat
            )

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        General wrapper for steam ejector initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state blocks
        self.properties_motive_steam.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl, state_args=state_args
        )
        init_log.info_high("Motive steam inlet initialization complete")

        self.properties_entrained_vapor.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl, state_args=state_args
        )
        init_log.info_high("Entrained vapor inlet initialization complete")

        state_args_discharge = {}
        state_dict_motive = self.properties_motive_steam[
            self.flowsheet().config.time.first()
        ].define_port_members()
        state_dict_entrained = self.properties_entrained_vapor[
            self.flowsheet().config.time.first()
        ].define_port_members()

        for k in state_dict_motive.keys():
            if state_dict_motive[k].is_indexed():
                state_args_discharge[k] = {}
                for m in state_dict_motive[k].keys():
                    state_args_discharge[k][m] = (
                        state_dict_motive[k][m].value + state_dict_entrained[k][m].value
                    )
            else:
                state_args_discharge[k] = 0.5 * (
                    state_dict_motive[k].value + state_dict_entrained[k].value
                )

        self.properties_discharge_mix.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl, state_args=state_args_discharge
        )
        init_log.info_high("Discharge mix initialization complete")

        if degrees_of_freedom(self) != 0:
            raise Exception(
                f"{self.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(self)}"
            )

        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high(
            "Initialization steam ejector {}.".format(idaeslog.condition(res))
        )

        # ---------------------------------------------------------------------
        # Release state
        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "Entrainment ratio": self.entrainment_ratio,
            "Compression ratio": self.compression_ratio,
        }

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Motive steam": self.inlet_motive_steam,
                "inlet_entrained_vapor": self.inlet_entrained_vapor,
                "outlet_discharge_mix": self.outlet_discharge_mix,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

    @property
    def default_costing_method(self):
        return cost_steam_ejector
